import itertools
from typing import Dict, List

from langchain_core.language_models import BaseChatModel
from langchain_core.runnables import RunnableSerializable, chain
from more_itertools import chunked

from aidial_rag_eval.generation.models.lambdas import json_to_list
from aidial_rag_eval.generation.models.refusal_detectors.base_refusal_detector import (
    RefusalDetector,
)
from aidial_rag_eval.generation.models.refusal_detectors.refusal_template import (
    refusal_prompt,
)
from aidial_rag_eval.generation.types import RefusalReturn
from aidial_rag_eval.generation.utils.progress_bar import ProgressBarCallback
from aidial_rag_eval.types import Answer


@chain
def returns_to_refusal_return(input_: List) -> List[RefusalReturn]:
    """
    The final part of the chain, which calculates answer refusals
    for each answer in the batch based on the JSON output from the LLM.

    Parameters
    -----------
    input_: List
        Output from the LLM, where each batch element is tagged with "REJ"
        if the answer is an answer refusal, or "ANS" otherwise.

    Returns
    ------------
    List[RefusalReturn]
        Returns a list of RefusalReturn, where each input answer from the batch
        is assigned a 1. if it is an answer refusal, or 0. otherwise.
    """
    return [RefusalReturn(refusal=float(tag == "REJ")) for tag in input_]


@chain
def wrap_answers(input_: Dict) -> Dict:
    assert type(input_) is dict
    return {
        "answers": [
            f"<answer{index + 1}> {hypothesis} </answer{index + 1}>"
            for index, hypothesis in enumerate(input_["answers"])
        ],
    }


class LLMRefusalDetector(RefusalDetector):
    """
    Experimental:
    The LLMRefusalDetector is designed to calculate
    answer refusal using a LLM.
    """

    _chain: RunnableSerializable
    """A chain that contains the core logic, which includes:
    the prompt, model, conversion of output content to JSON,
    and  and transformation of JSON into RefusalReturn."""

    max_concurrency: int
    """Configuration attribute for _chain.batch,
    indicating how many prompts will be sent in parallel."""

    def __init__(
        self,
        model: BaseChatModel,
        max_concurrency: int,
    ):

        self._chain = (
            wrap_answers
            | refusal_prompt
            | model
            | json_to_list
            | returns_to_refusal_return
        )
        self.max_concurrency = max_concurrency

    def get_refusal(
        self, answers: List[Answer], show_progress_bar: bool
    ) -> List[RefusalReturn]:
        """
        Method that calls a chain to calculate answer refusals
        for each answer.

        Parameters
        -----------
        answers: List[Answer] : List[InferenceInputs]
            A list of answers or ground truth answers.

        show_progress_bar : bool
            A flag that controls the display of a progress bar

        Returns
        ------------
        List[RefusalReturn]
            Returns the answer refusal for each answer.
        """
        batches = list(chunked(answers, 10))

        with ProgressBarCallback(len(batches), show_progress_bar) as cb:
            refusal_returns = self._chain.batch(
                [
                    {
                        "answers": hypotheses,
                    }
                    for hypotheses in batches
                ],
                config={"callbacks": [cb], "max_concurrency": self.max_concurrency},
            )
        refusal_returns = [
            (
                refusal_return
                if len(refusal_return) == len(input_batch)
                else [RefusalReturn(refusal=0.0)] * len(input_batch)
            )
            for refusal_return, input_batch in zip(refusal_returns, batches)
        ]
        return list(itertools.chain.from_iterable(refusal_returns))



from dataclasses import dataclass
from enum import Enum
from typing import List, TypeVar

FactType = TypeVar("FactType")
ContextChunk = str
Document = str
Text = str

Context = List[ContextChunk]
Facts = List[FactType]
Documents = List[Document]
Question = Text
Answer = Text
GroundTruthAnswer = Text


@dataclass
class GroundTruth:
    question: Question
    documents: Documents
    facts: Facts
    answer: GroundTruthAnswer


@dataclass
class CollectedAnswers:
    question: Question
    documents: Documents
    context: Context
    answer: Answer


class GroundTruthColumns(str, Enum):
    QUESTION = "question"
    DOCUMENTS = "documents"
    FACTS = "facts"
    ANSWER = "answer"


class AnswerColumns(str, Enum):
    QUESTION = "question"
    DOCUMENTS = "documents"
    CONTEXT = "context"
    ANSWER = "answer"


class MergedColumns(str, Enum):
    QUESTION = GroundTruthColumns.QUESTION.value
    CONTEXT = AnswerColumns.CONTEXT.value
    ANSWER = AnswerColumns.ANSWER.value
    GROUND_TRUTH_ANSWER = "ground_truth_answer"
    DOCUMENTS = GroundTruthColumns.DOCUMENTS.value
    FACTS = GroundTruthColumns.FACTS.value
    JOINED_CONTEXT = "joined_context"


MERGED_KEY_COLUMNS = [MergedColumns.DOCUMENTS, MergedColumns.QUESTION]
ANSWERS_COLUMNS = [col.value for col in AnswerColumns if isinstance(col, AnswerColumns)]
GROUND_TRUTH_COLUMNS = [
    col.value for col in GroundTruthColumns if isinstance(col, GroundTruthColumns)
]



from typing import Dict, List

from langchain_core.language_models import BaseChatModel
from langchain_core.runnables import RunnablePassthrough, RunnableSerializable, chain

from aidial_rag_eval.generation.models.lambdas import json_to_list
from aidial_rag_eval.generation.models.statement_extractor.base_statement_extractor import (
    StatementExtractor,
)
from aidial_rag_eval.generation.models.statement_extractor.statement_extractor_template import (
    statement_prompt,
)
from aidial_rag_eval.generation.types import HypothesisSegment, Statement
from aidial_rag_eval.generation.utils.progress_bar import ProgressBarCallback


@chain
def list_to_statements(
    llm_outputs_with_inputs: Dict,
) -> List[List[str]]:
    """
    Function is part of a chain that extracts segments from a list.

    Parameters
    -----------
    llm_outputs_with_inputs : Dict
        Passed inputs with output list of dicts from the LLM with extracted statements.

    Returns
    ------------
    List[List[str]]
        The extracted statements. if the LLM output is valid;
        otherwise, return an original list of hypothesis_segments,
        each hypothesis_segment is wrapped.
    """
    try:
        statements_for_hypothesis_segments = llm_outputs_with_inputs[
            "llm_output_statements"
        ]
        hypothesis_segments = llm_outputs_with_inputs["hypothesis_segments"]
        assert len(hypothesis_segments) == len(statements_for_hypothesis_segments)
        return [
            return_dict["statements"]
            for return_dict in llm_outputs_with_inputs["llm_output_statements"]
        ]
    except (
        TypeError,
        KeyError,
        AssertionError,
    ):
        return [
            [hypothesis_segment]
            for hypothesis_segment in llm_outputs_with_inputs["hypothesis_segments"]
        ]


@chain
def wrap_hypotheses(input_: Dict) -> Dict:
    assert type(input_) is dict
    return {
        "hypotheses": [
            f"<hypothesis{index + 1}> {hypothesis_segment} </hypothesis{index + 1}>"
            for index, hypothesis_segment in enumerate(input_["hypothesis_segments"])
        ],
    }


class LLMStatementExtractor(StatementExtractor):
    """
    The LLMStatementExtractor is designed to extract
    statements from a hypothesis segment using a LLM.
    """

    _chain: RunnableSerializable
    """A chain that contains the core logic, which includes:
    the prompt, model, conversion of output content to JSON,
    and transformation of JSON into statements."""

    max_concurrency: int
    """Configuration attribute for _chain.batch,
    indicating how many prompts will be sent in parallel."""

    def __init__(
        self,
        model: BaseChatModel,
        max_concurrency: int,
    ):

        self._chain = (
            RunnablePassthrough.assign(
                llm_output_statements=wrap_hypotheses
                | statement_prompt
                | model
                | json_to_list
            )
            | list_to_statements
        )
        self.max_concurrency = max_concurrency

    def extract(
        self,
        list_of_hypothesis_segments: List[List[HypothesisSegment]],
        show_progress_bar: bool,
    ) -> List[List[List[Statement]]]:
        """
        Method that calls a chain to extract statements from each
        hypothesis segment.

        Parameters
        -----------
        list_of_hypothesis_segments : List[List[HypothesisSegment]]
            A list of hypothesis segments as a sources of statements.

        show_progress_bar : bool
            A flag that controls the display of a progress bar

        Returns
        ------------
        List[List[List[Statement]]]
            Returns the statements for each hypothesis segment.
        """

        with ProgressBarCallback(
            len(list_of_hypothesis_segments), show_progress_bar
        ) as cb:
            returns = self._chain.batch(
                [
                    {
                        "hypothesis_segments": hypothesis_segments,
                    }
                    for hypothesis_segments in list_of_hypothesis_segments
                ],
                config={"callbacks": [cb], "max_concurrency": self.max_concurrency},
            )
        return returns



from typing import Any, Callable, Dict, List, Optional

import pandas as pd
from langchain_core.language_models import BaseChatModel

from aidial_rag_eval.generation.inference import calculate_batch_inference
from aidial_rag_eval.generation.refusal import calculate_batch_refusal
from aidial_rag_eval.generation.types import MetricBind
from aidial_rag_eval.types import MergedColumns

C2A_INFERENCE_PREFIX = "ctx_ans_"
A2GT_INFERENCE_PREFIX = "ans_gt_"
GT2A_INFERENCE_PREFIX = "gt_ans_"

ANSWER_REFUSAL_PREFIX = "answer_"
GT_ANSWER_REFUSAL_PREFIX = "ground_truth_"


def _get_column_as_list_str(dataframe: pd.DataFrame, column: str) -> List[Any]:
    list_str = dataframe[column].to_list()
    assert isinstance(list_str, list)
    return list_str


def _wrapped_dataframe_inference(
    df_merged: pd.DataFrame,
    premise_column: str,
    hypothesis_column: str,
    llm: BaseChatModel,
    prefix: str,
    question_column: Optional[str] = None,
    document_column: Optional[str] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> pd.DataFrame:
    inference_returns = calculate_batch_inference(
        premises=_get_column_as_list_str(df_merged, premise_column),
        hypotheses=_get_column_as_list_str(df_merged, hypothesis_column),
        llm=llm,
        questions=(
            _get_column_as_list_str(df_merged, question_column)
            if question_column is not None
            else None
        ),
        list_documents=(
            _get_column_as_list_str(df_merged, document_column)
            if document_column is not None
            else None
        ),
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    return pd.DataFrame(
        [vars(inference_return) for inference_return in inference_returns]
    ).add_prefix(prefix)


def _wrapped_dataframe_refusal(
    df_merged: pd.DataFrame,
    answer_column: str,
    llm: BaseChatModel,
    prefix: str,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> pd.DataFrame:
    refusal_returns = calculate_batch_refusal(
        answers=_get_column_as_list_str(df_merged, answer_column),
        llm=llm,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    return pd.DataFrame([vars(refusal) for refusal in refusal_returns]).add_prefix(
        prefix
    )


def context_to_answer_inference(
    df_merged, llm, max_concurrency, show_progress_bar, **kwargs
) -> pd.DataFrame:
    return _wrapped_dataframe_inference(
        df_merged=df_merged,
        premise_column=MergedColumns.JOINED_CONTEXT,
        hypothesis_column=MergedColumns.ANSWER,
        llm=llm,
        prefix=C2A_INFERENCE_PREFIX,
        # The last segment(sentence) of the question is attached to the premise.
        # This can be helpful when the premise is the answer or ground truth
        # and it is simple. Example:
        # question: how many boxes are in the cupboard?
        # answer (premise): 3.
        # When the premise is the context, the question is not needed.
        question_column=None,
        document_column=MergedColumns.DOCUMENTS,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )


def answer_to_ground_truth_inference(
    df_merged, llm, max_concurrency, show_progress_bar, **kwargs
) -> pd.DataFrame:
    return _wrapped_dataframe_inference(
        df_merged=df_merged,
        premise_column=MergedColumns.ANSWER,
        hypothesis_column=MergedColumns.GROUND_TRUTH_ANSWER,
        llm=llm,
        prefix=A2GT_INFERENCE_PREFIX,
        question_column=MergedColumns.QUESTION,
        document_column=MergedColumns.DOCUMENTS,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )


def ground_truth_to_answer_inference(
    df_merged, llm, max_concurrency, show_progress_bar, **kwargs
) -> pd.DataFrame:
    return _wrapped_dataframe_inference(
        df_merged=df_merged,
        premise_column=MergedColumns.GROUND_TRUTH_ANSWER,
        hypothesis_column=MergedColumns.ANSWER,
        llm=llm,
        prefix=GT2A_INFERENCE_PREFIX,
        question_column=MergedColumns.QUESTION,
        document_column=MergedColumns.DOCUMENTS,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )


def answer_refusal(
    df_merged, llm, max_concurrency, show_progress_bar, **kwargs
) -> pd.DataFrame:
    return _wrapped_dataframe_refusal(
        df_merged=df_merged,
        answer_column=MergedColumns.ANSWER,
        llm=llm,
        prefix=ANSWER_REFUSAL_PREFIX,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )


def ground_truth_refusal(
    df_merged, llm, max_concurrency, show_progress_bar, **kwargs
) -> pd.DataFrame:
    return _wrapped_dataframe_refusal(
        df_merged=df_merged,
        answer_column=MergedColumns.GROUND_TRUTH_ANSWER,
        llm=llm,
        prefix=GT_ANSWER_REFUSAL_PREFIX,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )


metric_binds_dict: Dict[MetricBind, Callable] = {
    "context_to_answer_inference": context_to_answer_inference,
    "answer_to_ground_truth_inference": answer_to_ground_truth_inference,
    "ground_truth_to_answer_inference": ground_truth_to_answer_inference,
    "answer_refusal": answer_refusal,
    "ground_truth_refusal": ground_truth_refusal,
}
metric_bind_keys: List[MetricBind] = list(metric_binds_dict.keys())



from typing import Optional, Tuple

from aidial_rag_eval.facts.match_str_facts import match_str_facts
from aidial_rag_eval.retrieval.types import ContextChunk, FactMatchResult
from aidial_rag_eval.types import Context, Facts


class CitationMatcher:
    Fact = str

    @staticmethod
    def _canonize(text: str) -> tuple[str, list[int]]:
        text = text.replace("\t", " ")
        text = "".join([c.lower() if c.isalnum() else " " for c in text])

        new_text = ""
        num_skipped_chars_for_pos = []

        # skip duplicated spaces
        for i, c in enumerate(text):
            if c == " " and new_text[-1:] == " ":
                continue
            new_text += c
            num_skipped_chars_for_pos.append(i - len(new_text) + 1)

        num_skipped_chars_for_pos.append(len(text) - len(new_text))

        return new_text, num_skipped_chars_for_pos

    @staticmethod
    def _match(fact: Fact, context_chunk: ContextChunk) -> Optional[Tuple[int, int]]:
        canonized_fact, _ = CitationMatcher._canonize(fact)
        canonized_context_chunk, skipped_chars = CitationMatcher._canonize(
            context_chunk
        )

        start = canonized_context_chunk.find(canonized_fact)
        if start == -1:
            return None
        end = start + len(canonized_fact)

        start = start + skipped_chars[start]
        end = end + skipped_chars[end]
        return start, end

    @staticmethod
    def match_facts(facts: Facts[Fact], context: Context) -> FactMatchResult:
        return match_str_facts(facts, context, CitationMatcher._match)



import pandas as pd

from aidial_rag_eval.types import (
    ANSWERS_COLUMNS,
    GROUND_TRUTH_COLUMNS,
    MERGED_KEY_COLUMNS,
    AnswerColumns,
    GroundTruthColumns,
    MergedColumns,
)


def merge_ground_truth_and_answers(
    ground_truth: pd.DataFrame, answers: pd.DataFrame
) -> pd.DataFrame:
    """
    Merges ground_truth and collected answers dataframes.

    Parameters
    -----------
    ground_truth : pd.DataFrame
        contains the ground truth answers
        The structure of the ground truth
        is described in `aidial_rag_eval.types.GroundTruthAnswers`.

    answers : pd.DataFrame
        contains the collected answers
        The structure of the collected answers
        is described in `aidial_rag_eval.types.CollectedAnswers`.

    Returns
    ------------
    pd.DataFrame
        Returns merged ground_truth dataframe and answers dataframe.
        The structure of the df_merged
        is described in `aidial_rag_eval.types.MergedColumns`.
    """
    ground_truth_copy = ground_truth[
        ground_truth.columns.intersection(GROUND_TRUTH_COLUMNS)
    ].copy()
    assert isinstance(ground_truth_copy, pd.DataFrame)
    answers_copy = answers[answers.columns.intersection(ANSWERS_COLUMNS)].copy()
    assert isinstance(answers_copy, pd.DataFrame)
    ground_truth_copy = ground_truth_copy.rename(
        columns={
            GroundTruthColumns.ANSWER.value: MergedColumns.GROUND_TRUTH_ANSWER.value
        }
    )
    ground_truth_copy[GroundTruthColumns.DOCUMENTS] = ground_truth_copy[
        GroundTruthColumns.DOCUMENTS
    ].apply(frozenset)
    answers_copy[AnswerColumns.DOCUMENTS] = answers_copy[AnswerColumns.DOCUMENTS].apply(
        frozenset
    )
    data = pd.merge(
        ground_truth_copy,
        answers_copy,
        on=MERGED_KEY_COLUMNS,
    )
    data[MergedColumns.DOCUMENTS] = answers_copy.loc[
        data.index, MergedColumns.DOCUMENTS
    ]
    data[MergedColumns.DOCUMENTS] = data[MergedColumns.DOCUMENTS].apply(list)
    return data



from abc import ABC, abstractmethod
from typing import List

from aidial_rag_eval.generation.types import RefusalReturn


class RefusalDetector(ABC):
    """
    Abstract base class for creating RefusalDetector to calculate
    answer refusal.
    """

    @abstractmethod
    def get_refusal(
        self, answers: List[str], show_progress_bar: bool
    ) -> List[RefusalReturn]:
        return [RefusalReturn(refusal=0.0)] * len(answers)



# flake8: noqa
from langchain_core.prompts import PromptTemplate

decontextualization_template = """
The task is to replace all pronouns in a segments with their corresponding nouns or proper names when their referents are known.
You will receive segments.
If a segment is nonsensical, a reference, link, or meaningless, return it unchanged.
If unsure what to do with segment, return the original segment.
Only perform the task; do not shorten, simplify, or correct errors.
Do not provide explanations.

For example: "My mom is a good person.", "She always takes care of me." you must return:
```json
{
    "segments": [
        "My mom is a good person.",
        "My mom always takes care of me.",
        ...
    ]
}
```

Your response template:
```json
{
    "segments": [
        << Segment 1 >>,
        << Segment 2 >>,
        ...
    ]
}
```
List of input segments:
```json
{
    "segments": {{ sentences_str }}
}
```
Important: before generating response, check the number and structure of segments. The response must have the same number of segments, split the same way.
"""

decontextualization_prompt = PromptTemplate.from_template(
    template=decontextualization_template,
    template_format="jinja2",
)



from json import JSONDecodeError
from typing import Dict, List

from langchain_core.exceptions import OutputParserException
from langchain_core.messages import AIMessage
from langchain_core.runnables import chain
from langchain_core.utils.json import parse_json_markdown


@chain
def json_to_list(input_: AIMessage) -> List:
    """
    Intermediate part of the chain converts the output of the LLM into a JSON,
    where the outer structure is a dictionary containing a list.

    Parameters
    -----------
    input_: AIMessage
        Output from the LLM.

    Returns
    ------------
    List
        Returns a list; if the LLM output was incorrect, returns an empty list.
    """
    try:
        return_dict = parse_json_markdown(str(input_.content))
        assert isinstance(return_dict, dict)
        return_list = return_dict[list(return_dict.keys())[0]]
        assert isinstance(return_list, list)
        return return_list
    except (IndexError, OutputParserException, JSONDecodeError, AssertionError):
        return []


@chain
def json_to_dict(input_: AIMessage) -> Dict[str, List[str]]:
    """
    Intermediate part of the chain that converts the output of the LLM into a JSON,
    the outer part of which is a dict.

    Parameters
    -----------
    input_ : AIMessage
        The output from the LLM.

    Returns
    ------------
    Dict[str, List[str]]
        LLM output if valid;
        otherwise, an empty dict is returned.
    """
    try:
        return_dict = parse_json_markdown(str(input_.content))
        assert isinstance(return_dict, dict)
        return return_dict
    except (
        TypeError,
        OutputParserException,
        JSONDecodeError,
        AssertionError,
    ):
        return {}



from typing import Optional, Tuple

from aidial_rag_eval.facts.match_str_facts import match_str_facts
from aidial_rag_eval.retrieval.types import ContextChunk, FactMatchResult
from aidial_rag_eval.types import Context, Facts


class ExactStringMatcher:
    Fact = str

    @staticmethod
    def _match(fact: Fact, context_chunk: ContextChunk) -> Optional[Tuple[int, int]]:
        if fact == context_chunk:
            return (0, len(context_chunk))
        return None

    @staticmethod
    def match_facts(facts: Facts[Fact], context: Context) -> FactMatchResult:
        return match_str_facts(facts, context, ExactStringMatcher._match)



from datetime import datetime
from typing import Optional, Union

import fsspec
import pandas as pd
from pydantic import Field, TypeAdapter
from pydantic.dataclasses import dataclass


@dataclass
class DatasetMetadata:
    created_at: Optional[datetime]
    sources: list["Dataset"]
    tools: dict[str, str]
    file_info: dict = Field(default_factory=dict)
    metrics: dict = Field(default_factory=dict)
    statistics: dict = Field(default_factory=dict)

    @staticmethod
    def read_metadata(metadata_path: str) -> "DatasetMetadata":
        fs, _ = fsspec.url_to_fs(metadata_path)
        return TypeAdapter(DatasetMetadata).validate_json(fs.read_bytes(metadata_path))

    @staticmethod
    def create_missing_metadata(data_path: str) -> "DatasetMetadata":
        fs, _ = fsspec.url_to_fs(data_path)
        file_info = fs.info(data_path)
        return DatasetMetadata(None, [], {}, file_info, {}, {})


@dataclass
class Dataset:
    metadata_path: str
    data_path: str
    metadata: DatasetMetadata

    @staticmethod
    def handle_path(path: str) -> tuple[str, str]:
        if path.endswith(".parquet.metadata.json"):
            return path, path[: -len(".metadata.json")]
        if path.endswith(".parquet"):
            return path + ".metadata.json", path
        raise ValueError(f"Invalid path: {path}")

    @staticmethod
    def from_metadata(path: str):
        metadata_path, data_path = Dataset.handle_path(path)

        fs, _ = fsspec.url_to_fs(data_path)
        if fs.exists(metadata_path):
            metadata = DatasetMetadata.read_metadata(metadata_path)
        else:
            metadata = DatasetMetadata.create_missing_metadata(data_path)

        return Dataset(metadata_path, data_path, metadata)

    def read_dataframe(self, *args, **kwargs) -> pd.DataFrame:
        return pd.read_parquet(self.data_path, *args, **kwargs)

    @staticmethod
    def write_dataframe(
        df: pd.DataFrame,
        path: str,
        sources: Optional[list["Dataset"]] = None,
        tools: Optional[dict[str, str]] = None,
        metrics: Optional[dict[str, float]] = None,
        statistics: Optional[dict[str, float]] = None,
        *args,
        **kwargs,
    ):
        sources = sources or []
        tools = tools or {}
        metrics = metrics or {}
        statistics = statistics or {}
        metadata_path, data_path = Dataset.handle_path(path)

        fs, _ = fsspec.url_to_fs(data_path)
        df.to_parquet(data_path, *args, **kwargs)
        file_info: dict = fs.info(data_path)

        metadata = DatasetMetadata(
            datetime.now(),
            sources,
            tools,
            file_info,
            metrics,
            statistics,
        )
        fs.write_bytes(
            metadata_path, TypeAdapter(DatasetMetadata).dump_json(metadata, indent=2)
        )

        return Dataset(metadata_path, data_path, metadata)


def source_dataset(dataset: Union[str, Dataset]):
    if isinstance(dataset, str):
        return Dataset.from_metadata(dataset)
    return dataset



import json
from json import JSONDecodeError
from typing import Dict, List

from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import AIMessage
from langchain_core.runnables import RunnableSerializable, chain
from langchain_core.utils.json import parse_json_markdown

from aidial_rag_eval.generation.models.converters.base_converter import SegmentConverter
from aidial_rag_eval.generation.models.converters.decontextualization_template import (
    decontextualization_prompt,
)
from aidial_rag_eval.generation.types import TextSegment
from aidial_rag_eval.generation.utils.progress_bar import ProgressBarCallback
from aidial_rag_eval.generation.utils.segmented_text import SegmentedText


@chain
def json_to_dict_segments(input_: AIMessage) -> List[str]:
    """
    Function is part of a chain that extracts segments from an AIMessage.

    Parameters
    -----------
    input_ : AIMessage
        The output from the LLM which includes content with transformed segments.

    Returns
    ------------
    List[str]
        The transformed segments if the LLM output is valid;
        otherwise, an empty list is returned.
    """
    try:
        return_dict = parse_json_markdown(str(input_.content))
        assert isinstance(return_dict, dict)
        return return_dict["segments"]
    except (
        TypeError,
        KeyError,
        OutputParserException,
        JSONDecodeError,
        AssertionError,
    ):
        return []


@chain
def sentences_to_json_list(input_: Dict) -> Dict:
    assert type(input_) is dict
    return {"sentences_str": json.dumps(input_["sentences"])}


class LLMNoPronounsConverter(SegmentConverter):
    """
    The LLMNoPronounsBatchConverter is designed to replace pronouns
    in text segments using a LLM.

    Input is a list of SegmentedText objects.
    If a SegmentedText object contains more than one segment,
    segments are sent in a prompt to the LLM.
    In a prompt, the first segment is used only for context,
    and pronoun replacement is performed only in the remaining segments.
    """

    _chain: RunnableSerializable
    """A chain that contains the core logic, which includes:
    the prompt, model, conversion of output content to JSON,
    and extraction of segments from JSON."""

    max_concurrency: int
    """Configuration attribute for _chain.batch,
    indicating how many prompts will be sent in parallel."""

    def __init__(
        self,
        model: BaseChatModel,
        max_concurrency: int,
    ):

        self._chain = (
            sentences_to_json_list
            | decontextualization_prompt
            | model
            | json_to_dict_segments
        )
        self.max_concurrency = max_concurrency

    def transform_texts(
        self, segmented_texts: List[SegmentedText], show_progress_bar: bool
    ):
        """
        Method that converts segmented texts by replacing pronouns using an LLM.
        The LLM processes segments,
        where the additional first segment is not converted
        but is provided for context to enable the conversion of the second sentence.
        The LLM returns converted segments.
        If the invariant of the length of input and output segment batches
        is not maintained, the segments of this batch are not replaced.

        Parameters
        -----------
        segmented_texts : List[SegmentedText]
            A list of segmented texts where segment replacement occurs.

        show_progress_bar : bool
            A flag that controls the display of a progress bar.
        """
        original_segment_batches: List[List[TextSegment]] = []
        segment_ids: List[int] = []
        for text_id, segmented_text in enumerate(segmented_texts):
            segments = segmented_text.segments
            if len(segments) <= 1:
                continue
            original_segment_batches.append(segments)
            segment_ids.append(text_id)

        no_pronouns_segment_batches = self._get_no_pronouns_segments(
            original_segment_batches, show_progress_bar
        )

        for text_id, no_pronouns_segment_batch, original_segment_batch in zip(
            segment_ids, no_pronouns_segment_batches, original_segment_batches
        ):
            if len(no_pronouns_segment_batch) != len(original_segment_batch):
                continue
            segmented_texts[text_id].replace_segments(
                no_pronouns_segment_batch[1:],
                1,
            )

    def _get_no_pronouns_segments(
        self,
        original_segment_batches: List[List[TextSegment]],
        show_progress_bar: bool,
    ) -> List[List[TextSegment]]:
        """
        Method that calls _chain to replace pronouns.

        Parameters
        -----------
        original_segment_batches : List[List[str]]
            Segments of texts.

        show_progress_bar : bool
            A flag that controls the display of a progress bar.
        Returns
        ------------
        List[List[str]]
            List of converted segments, divided into batches.
        """
        with ProgressBarCallback(
            len(original_segment_batches), show_progress_bar
        ) as cb:
            no_pronouns_segment_batches = self._chain.batch(
                [
                    {
                        "sentences": batch,
                    }
                    for batch in original_segment_batches
                ],
                config={"callbacks": [cb], "max_concurrency": self.max_concurrency},
            )
        return no_pronouns_segment_batches



from abc import ABC, abstractmethod
from typing import List

from aidial_rag_eval.generation.types import InferenceInputs, InferenceScore


class InferenceScorer(ABC):
    """
    Abstract base class for creating InferenceScorer to calculate
    inference of a hypothesis from a premise.
    """

    @abstractmethod
    def get_inference(
        self,
        inference_inputs: List[InferenceInputs],
        show_progress_bar: bool,
    ) -> List[InferenceScore]:
        return [InferenceScore(inference=0.0, explanation="")] * len(inference_inputs)



import functools

import numpy as np

from aidial_rag_eval.retrieval.types import (
    ContextRelevance,
    FactMatchResult,
    FactsRanks,
)


def calculate_recall(facts_ranks: FactsRanks) -> np.float64:
    total_facts = len(facts_ranks)
    relevant_facts = np.count_nonzero(facts_ranks >= 0)
    recall = np.float64(relevant_facts) / np.float64(total_facts)
    return recall


def calculate_precision(context_relevance: ContextRelevance) -> np.float64:
    total_context = len(context_relevance)
    if total_context == 0:
        return np.float64(0.0)
    relevant_context = np.count_nonzero(context_relevance > 0)
    precision = np.float64(relevant_context) / np.float64(total_context)
    return precision


def calculate_mrr(facts_ranks: FactsRanks) -> np.floating:
    ranks = facts_ranks.astype(float) + 1
    ranks[ranks == 0] = np.inf
    mrr = np.mean(1 / ranks)
    return mrr


def calculate_f1(precision: np.float64, recall: np.float64, **kwargs) -> np.float64:
    if precision == 0 and recall == 0:
        return np.float64(0.0)
    else:
        return (
            np.float64(2.0)
            * np.float64(precision * recall)
            / np.float64(precision + recall)
        )


def wrap_facts_metric(metric, fields=FactMatchResult._fields):
    @functools.wraps(metric)
    def wrapped_metric(row: FactMatchResult):
        return metric(*[getattr(row, field) for field in fields])

    return wrapped_metric


METRICS_CONFIG = {
    "recall": wrap_facts_metric(calculate_recall, ["facts_ranks"]),
    "precision": wrap_facts_metric(calculate_precision, ["context_relevance"]),
    "mrr": wrap_facts_metric(calculate_mrr, ["facts_ranks"]),
}

SECONDARY_METRICS_CONFIG = {
    "f1": calculate_f1,
}


def calculate_metrics(row: FactMatchResult) -> dict:
    result = {name: metric(row) for name, metric in METRICS_CONFIG.items()}
    result.update(
        {name: metric(**result) for name, metric in SECONDARY_METRICS_CONFIG.items()}
    )
    return result



from aidial_rag_eval.dataframe.metrics import (
    calculate_generation_metrics,
    create_rag_eval_metrics_report,
    create_retrieval_metrics_report,
)

__all__ = [
    "create_rag_eval_metrics_report",
    "create_retrieval_metrics_report",
    "calculate_generation_metrics",
]



from typing import Any, List, Optional, Union

import pandas as pd
from langchain_core.language_models import BaseChatModel

from aidial_rag_eval.dataframe import create_rag_eval_metrics_report
from aidial_rag_eval.dataframe.match_facts import DEFAULT_MATCHER
from aidial_rag_eval.dataset import Dataset, source_dataset
from aidial_rag_eval.generation.types import MetricBind
from aidial_rag_eval.retrieval.types import Matcher
from aidial_rag_eval.utils import get_tools_versions


def evaluate(
    ground_truth: Union[str, Dataset],
    answers: Union[str, Dataset],
    dest: str,
    matcher: Matcher = DEFAULT_MATCHER,
    llm: Optional[BaseChatModel] = None,
    fs: Any = None,
    metric_binds: Optional[List[MetricBind]] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> Dataset:
    """
    Calculates RAG evaluation metrics from input
    Parquet datasets and writes the results to the specified destination.

    Parameters
    -----------
    ground_truth : Union[str, Dataset]
        Path to the Parquet file containing ground truth answers or
        a Dataset object specifying the Parquet file with ground truth answers.
        The structure of the ground truth Parquet file
        is described in `aidial_rag_eval.types.GroundTruthAnswers`.

    answers : Union[str, Dataset]
        Path to the Parquet file containing collected answers or
        a Dataset object specifying the Parquet file with collected answers.
        The structure of the collected answers Parquet file
        is described in `aidial_rag_eval.types.CollectedAnswers`.

    dest : str
        Path for the output file in Parquet format.

    matcher : Matcher, default=DEFAULT_MATCHER
        An object responsible for matching facts
        from the ground truth to the context in the answers.

    llm : BaseChatModel, optional, default=None
        A Langchain chat model used for calculating generation metrics.

    fs : Any, optional, default=None
        File system to be used.

    metric_binds : List[MetricBind], optional, default=None
        A list of string constants from `aidial_rag_eval.metric_binds`
        specifying the generation metrics.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        To display a progress bar during LLM requests.

    Returns
    ------------
    Dataset
        Returns global statistics for each metric.
    """
    ground_truth_dataset = source_dataset(ground_truth)
    answers_dataset = source_dataset(answers)

    ground_truth_df = ground_truth_dataset.read_dataframe(filesystem=fs)
    answers_df = answers_dataset.read_dataframe(filesystem=fs)

    df_final = create_rag_eval_metrics_report(
        ground_truth_df,
        answers_df,
        matcher,
        llm,
        metric_binds=metric_binds,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    aggregated_metrics = df_final.mean(numeric_only=True)
    assert isinstance(aggregated_metrics, pd.Series)
    aggregated_metrics.dropna(inplace=True)

    metrics = Dataset.write_dataframe(
        df_final,
        dest,
        sources=[ground_truth_dataset, answers_dataset],
        tools=get_tools_versions(),
        metrics=aggregated_metrics.to_dict(),
        statistics={
            "Ground truth size": len(ground_truth_df),
            "Answers size": len(answers_df),
            "Evaluation data size": len(df_final),
        },
        filesystem=fs,
    )
    return metrics



from importlib.metadata import version
from typing import Optional

PKG_NAME = "aidial-rag-eval"


def get_tools_versions(tools: Optional[list[str]] = None) -> dict[str, str]:
    tools = tools or [PKG_NAME]
    return {tool: version(tool) for tool in tools}



from typing import List

from langchain_core.language_models import BaseChatModel

from aidial_rag_eval.generation.models.refusal_detectors.llm_refusal_detector import (
    LLMRefusalDetector,
)
from aidial_rag_eval.generation.types import RefusalReturn
from aidial_rag_eval.generation.utils.segmented_text import SegmentedText
from aidial_rag_eval.types import Answer


def calculate_batch_refusal(
    answers: List[Answer],
    llm: BaseChatModel,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> List[RefusalReturn]:
    """
    Checks if the answers are answer refusal.

    Parameters
    -----------

        answers : List[str]
            The list of the answers.

        llm : BaseChatModel
            The Langchain chat model used for calculating inference.

        max_concurrency : int, default=8
            The maximum number of concurrent requests to the LLM.

        show_progress_bar : bool, default=True
            Whether to display a progress bar during LLM requests.

    Returns
    ------------
    RefusalReturn
        Returns the list of the answer refusals.
    """
    detector = LLMRefusalDetector(llm, max_concurrency)
    answers_split = [SegmentedText.from_text(text=answer) for answer in answers]
    # As a heuristic, we send only the first 3 segments in the prompt.
    # We believe that if there are 3 whole segments with information
    # that is not related to refusal to answer,
    # we will not consider such a response as a refusal to answer
    # in any case.
    first_answers_sentences = [
        answers_split[i].get_joined_segments_by_range(0, 3)
        for i in range(len(answers_split))
    ]
    if show_progress_bar:
        print("Getting refusal...")
    return detector.get_refusal(first_answers_sentences, show_progress_bar)


def calculate_refusal(
    answer: Answer,
    llm: BaseChatModel,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> RefusalReturn:
    """
    Checks if the answer is answer refusal.

    Parameters
    -----------

        answer : str
            The text of the answer.

        llm : BaseChatModel
            The Langchain chat model used for calculating inference.

        max_concurrency : int, default=8
            The maximum number of concurrent requests to the LLM.

        show_progress_bar : bool, default=True
            Whether to display a progress bar during LLM requests.

    Returns
    ------------
    RefusalReturn
        Returns the answer refusal.
    """
    refusal_returns = calculate_batch_refusal(
        answers=[answer],
        llm=llm,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    return refusal_returns[0]



from typing import Callable, List, Optional, Tuple

import nltk
from more_itertools import chunked

from aidial_rag_eval.generation.types import Text, TextSegment

Splitter = Callable[[TextSegment], List[TextSegment]]
SegmentChecker = Callable[[TextSegment], bool]


def get_delimiters(text: Text, segmented_text: List[TextSegment]) -> List[str]:
    delimiters = []
    start = 0
    for i, part in enumerate(segmented_text[:-1]):
        start += len(part)
        if start < len(text):
            delimiter = ""
            while start < len(text) and not text[start:].startswith(
                segmented_text[i + 1]
            ):
                delimiter += text[start]
                start += 1
            delimiters.append(delimiter)
    return delimiters


def join_with_delimiters(
    segmented_text: List[TextSegment], delimiters: List[str]
) -> Text:
    result = []
    for i in range(len(segmented_text)):
        result.append(segmented_text[i])
        if i < len(delimiters):
            result.append(delimiters[i])
    return "".join(result)


def _get_split_with_delimiters(
    original_segment: TextSegment, splitter: Splitter
) -> Tuple[List[TextSegment], List[str]]:
    segments = splitter(original_segment)
    if len(segments) > 1:
        delimiters = get_delimiters(original_segment, segments)
    else:
        delimiters = []
    assert len(segments) - 1 == len(delimiters)
    return segments, delimiters


def _default_checker(original_segment: TextSegment) -> bool:
    return True


class SegmentedText:
    """
    The SegmentedText class is designed to store
    the original text in a format that is divided into segments.

    Along with the segments, delimiters are also stored.
    When the segments are joined with the delimiters,
    the original text can be reconstructed, if the segments have not been replaced.
    """

    segments: List[TextSegment]
    """A list of segments into which the original text has been divided."""

    delimiters: List[str]
    """The gaps between segments.
    When segments are joined using these delimiters, they reconstruct the original text.
    len(delimiters) == len(segments) - 1."""

    def __init__(
        self,
        segments: List[TextSegment],
        delimiters: List[str],
    ):
        assert len(segments) - 1 == len(delimiters)
        self.segments = segments.copy()
        self.delimiters = delimiters.copy()

    @classmethod
    def from_text(cls, text: Text) -> "SegmentedText":
        nltk.download("punkt_tab", quiet=True)

        max_len = 500
        min_len = 10
        conditional_splitters: List[Splitter] = [
            lambda x: x.split("\n\n"),
            lambda x: x.split("\n"),
            lambda x: list(map(lambda y: "".join(y), chunked(x, max_len))),
        ]
        segmented_text = cls([text], [])
        segmented_text = apply_splitter_to_segmented_text(
            segmented_text, nltk.sent_tokenize
        )
        for splitter in conditional_splitters:
            segmented_text = apply_splitter_to_segmented_text(
                segmented_text, splitter, lambda x: len(x) > max_len
            )

        segmented_text = join_segments_to_next_by_condition(
            segmented_text, lambda x: len(x) < min_len
        )
        return segmented_text

    def replace_segments(self, segments: List[TextSegment], start_idx: int):
        self.segments[start_idx : start_idx + len(segments)] = segments

    def get_joined_segments_by_range(self, start: int, end: int) -> str:
        return join_with_delimiters(
            self.segments[start:end], self.delimiters[start : end - 1]
        )


def apply_splitter_to_segmented_text(
    segmented_text: SegmentedText,
    splitter: Splitter,
    checker: Optional[SegmentChecker] = None,
) -> SegmentedText:
    if checker is None:
        checker = _default_checker
    new_segments = []
    new_delimiters = []

    for i, original_segment in enumerate(segmented_text.segments):
        split_segment, split_delimiters = [original_segment], []
        if checker(original_segment):
            split_segment, split_delimiters = _get_split_with_delimiters(
                original_segment, splitter
            )
        new_segments.extend(split_segment)
        new_delimiters.extend(split_delimiters)
        if i < len(segmented_text.delimiters):
            new_delimiters.append(segmented_text.delimiters[i])
    return SegmentedText(new_segments, new_delimiters)


def join_segments_to_next_by_condition(
    segmented_text: SegmentedText, checker: SegmentChecker
) -> SegmentedText:
    if len(segmented_text.segments) <= 1:
        return SegmentedText(segmented_text.segments, segmented_text.delimiters)
    reversed_segments = [segmented_text.segments[-1]]
    reversed_delimiters = []

    # Iterate through self.segments in reverse.
    # If a segment meets a condition (e.g., too small),
    # join it with the previous segment in reversed_segments(next segment in self.segments).
    for i in reversed(range(len(segmented_text.segments) - 1)):
        if checker(segmented_text.segments[i]):
            reversed_segments[-1] = join_with_delimiters(
                [segmented_text.segments[i], reversed_segments[-1]],
                [segmented_text.delimiters[i]],
            )
        else:
            reversed_segments.append(segmented_text.segments[i])
            reversed_delimiters.append(segmented_text.delimiters[i])

    # Handle the last segment (the first one in the reversed_segments) separately.
    if len(reversed_segments) > 1 and checker(reversed_segments[0]):
        reversed_segments[0:2] = [
            join_with_delimiters(
                [reversed_segments[1], reversed_segments[0]],
                [reversed_delimiters[0]],
            )
        ]
        reversed_delimiters[0:1] = []
    return SegmentedText(reversed_segments[::-1], reversed_delimiters[::-1])



from abc import ABC, abstractmethod
from typing import List

from aidial_rag_eval.generation.utils.segmented_text import SegmentedText


class SegmentConverter(ABC):
    """
    Abstract base class for creating SegmentConverter.

    Input is a list of SegmentedText objects to convert.
    """

    @abstractmethod
    def transform_texts(
        self, segmented_texts: List[SegmentedText], show_progress_bar: bool
    ):
        pass



from typing import List, Optional

import pandas as pd
from langchain_core.language_models import BaseChatModel

from aidial_rag_eval.dataframe.match_facts import (
    DEFAULT_MATCHER,
    apply_matcher_to_merged_dataframe,
)
from aidial_rag_eval.dataframe.merge import merge_ground_truth_and_answers
from aidial_rag_eval.generation.metric_binds import metric_binds_dict
from aidial_rag_eval.generation.types import MetricBind, inference_column
from aidial_rag_eval.retrieval.metrics import (
    calculate_metrics as calculate_metrics_by_row,
)
from aidial_rag_eval.retrieval.types import Matcher
from aidial_rag_eval.types import MergedColumns


def apply_metrics_to_matched_results(match_result_data: pd.DataFrame) -> pd.DataFrame:
    retrieval_metrics = match_result_data.apply(
        calculate_metrics_by_row,
        axis=1,
        result_type="expand",
    )
    assert isinstance(retrieval_metrics, pd.DataFrame)
    return retrieval_metrics


def calculate_metrics(match_result_data: pd.DataFrame) -> pd.DataFrame:
    return pd.merge(
        match_result_data,
        apply_metrics_to_matched_results(match_result_data),
        left_index=True,
        right_index=True,
    )


def calculate_retrieval_metrics(
    df_merged: pd.DataFrame,
    matcher: Matcher = DEFAULT_MATCHER,
) -> pd.DataFrame:
    """
    Calculates RAG evaluation retrieval metrics from df_merged dataframe.

    Parameters
    -----------
    df_merged : pd.DataFrame
        merged ground truth answers and collected answers dataframes
        The structure of the df_merged
        is described in `aidial_rag_eval.types.MergedColumns`.

    matcher : Matcher, default=DEFAULT_MATCHER
        An object responsible for matching facts
        from the ground truth to the context in the answers.

    Returns
    ------------
    pd.DataFrame
        Returns retrieval metrics dataframe.
    """
    matched_result = apply_matcher_to_merged_dataframe(df_merged, matcher)
    matched_result_with_df = pd.merge(
        df_merged, matched_result, left_index=True, right_index=True
    )
    retrieval_metrics = apply_metrics_to_matched_results(matched_result_with_df)
    return pd.merge(
        matched_result, retrieval_metrics, left_index=True, right_index=True
    )


def create_retrieval_metrics_report(
    ground_truth: pd.DataFrame,
    answers: pd.DataFrame,
    matcher: Matcher = DEFAULT_MATCHER,
) -> pd.DataFrame:
    """
    Calculates RAG evaluation retrieval metrics from input dataframes.

    Parameters
    -----------
    ground_truth : pd.DataFrame
        contains the ground truth answers
        The structure of the ground truth
        is described in `aidial_rag_eval.types.GroundTruthAnswers`.

    answers : pd.DataFrame
        contains the collected answers
        The structure of the collected answers
        is described in `aidial_rag_eval.types.CollectedAnswers`.

    matcher : Matcher, default=DEFAULT_MATCHER
        An object responsible for matching facts
        from the ground truth to the context in the answers.

    Returns
    ------------
    pd.DataFrame
        Returns merged ground_truth dataframe,
        answers dataframe with retrieval metrics dataframe.
    """
    df_merged = merge_ground_truth_and_answers(ground_truth, answers)
    return pd.merge(
        df_merged,
        calculate_retrieval_metrics(df_merged, matcher),
        left_index=True,
        right_index=True,
    )


def calculate_generation_metrics(
    df_merged: pd.DataFrame,
    llm: BaseChatModel,
    metric_binds: List[MetricBind],
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> pd.DataFrame:
    """
    Calculates RAG evaluation generation metrics from df_merged dataframe.

    Parameters
    -----------
    df_merged : pd.DataFrame
        merged ground truth answers and collected answers dataframes
        The structure of the df_merged
        is described in `aidial_rag_eval.types.MergedColumns`.

    llm : BaseChatModel, optional, default=None
        A Langchain chat model used for calculating generation metrics.

    metric_binds : List[MetricBind], optional, default=None
        A list of string constants from `aidial_rag_eval.metric_binds`
        specifying the generation metrics.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        To display a progress bar during LLM requests.

    Returns
    ------------
    pd.DataFrame
        Returns generation metrics dataframe.
    """
    df_merged_copy = df_merged.copy()
    df_merged_copy[MergedColumns.JOINED_CONTEXT] = df_merged_copy[
        MergedColumns.CONTEXT
    ].apply(lambda x: "\n".join(x))
    metric_results = dict()
    for metric_bind in metric_binds:
        metric_results.update(
            metric_binds_dict[metric_bind](
                df_merged=df_merged_copy,
                llm=llm,
                max_concurrency=max_concurrency,
                show_progress_bar=show_progress_bar,
            ).to_dict(orient="series")
        )
    df_metrics = pd.DataFrame(data=metric_results)
    nli_columns = [
        column for column in metric_results.keys() if column.endswith(inference_column)
    ]
    if nli_columns:
        sub_df_nli = df_metrics[nli_columns]
        df_metrics["mean_" + inference_column] = sub_df_nli.mean(1)
        df_metrics["median_" + inference_column] = sub_df_nli.median(
            1
        )  # pyright: ignore # noqa

    return df_metrics


def create_generation_metrics_report(
    ground_truth: pd.DataFrame,
    answers: pd.DataFrame,
    llm: BaseChatModel,
    metric_binds: List[MetricBind],
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> pd.DataFrame:
    """
    Calculates RAG evaluation generation metrics from input dataframes.

    Parameters
    -----------
    ground_truth : pd.DataFrame
        contains the ground truth answers
        The structure of the ground truth
        is described in `aidial_rag_eval.types.GroundTruthAnswers`.

    answers : pd.DataFrame
        contains the collected answers
        The structure of the collected answers
        is described in `aidial_rag_eval.types.CollectedAnswers`.

    llm : BaseChatModel, optional, default=None
        A Langchain chat model used for calculating generation metrics.

    metric_binds : List[MetricBind], optional, default=None
        A list of string constants from `aidial_rag_eval.metric_binds`
        specifying the generation metrics.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        To display a progress bar during LLM requests.

    Returns
    ------------
    pd.DataFrame
        Returns merged ground_truth dataframe,
        answers dataframe with generation metrics dataframe.
    """
    df_merged = merge_ground_truth_and_answers(ground_truth, answers)
    df_metrics = calculate_generation_metrics(
        df_merged,
        llm,
        metric_binds,
        max_concurrency,
        show_progress_bar,
    )
    return pd.merge(df_merged, df_metrics, left_index=True, right_index=True)


def create_rag_eval_metrics_report(
    ground_truth: pd.DataFrame,
    answers: pd.DataFrame,
    matcher: Matcher = DEFAULT_MATCHER,
    llm: Optional[BaseChatModel] = None,
    metric_binds: Optional[List[MetricBind]] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> pd.DataFrame:
    """
    Calculates RAG evaluation metrics from input dataframes.

    Parameters
    -----------
    ground_truth : pd.DataFrame
        contains the ground truth answers
        The structure of the ground truth
        is described in `aidial_rag_eval.types.GroundTruthAnswers`.

    answers : pd.DataFrame
        contains the collected answers
        The structure of the collected answers
        is described in `aidial_rag_eval.types.CollectedAnswers`.

    matcher : Matcher, default=DEFAULT_MATCHER
        An object responsible for matching facts
        from the ground truth to the context in the answers.

    llm : BaseChatModel, optional, default=None
        A Langchain chat model used for calculating generation metrics.

    metric_binds : List[MetricBind], optional, default=None
        A list of string constants from `aidial_rag_eval.metric_binds`
        specifying the generation metrics.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        To display a progress bar during LLM requests.

    Returns
    ------------
    pd.DataFrame
        Returns merged ground_truth dataframe, answers dataframe with metrics dataframe.
    """
    df_merged = merge_ground_truth_and_answers(ground_truth, answers)
    retrieval_metrics = calculate_retrieval_metrics(df_merged, matcher)
    if not metric_binds:
        return pd.merge(df_merged, retrieval_metrics, left_index=True, right_index=True)
    if llm is None:
        raise ValueError("Argument 'llm' is required.")
    generation_metrics = calculate_generation_metrics(
        df_merged,
        llm,
        metric_binds,
        max_concurrency,
        show_progress_bar,
    )
    return pd.concat([df_merged, retrieval_metrics, generation_metrics], axis=1)



import json
from typing import Dict, List

import numpy as np
from langchain_core.language_models import BaseChatModel
from langchain_core.runnables import (
    RunnableBranch,
    RunnablePassthrough,
    RunnableSerializable,
    chain,
)

from aidial_rag_eval.generation.models.inference_scorers.base_inference_scorer import (
    InferenceScorer,
)
from aidial_rag_eval.generation.models.inference_scorers.inference_template import (
    inference_prompt,
)
from aidial_rag_eval.generation.models.lambdas import json_to_list
from aidial_rag_eval.generation.types import InferenceInputs, InferenceScore
from aidial_rag_eval.generation.utils.progress_bar import ProgressBarCallback


@chain
def returns_to_inference_score(llm_outputs_with_inputs: Dict) -> InferenceScore:
    """
    The final part of the chain for calculating inference.
    The inference is the average proportion of "ENT" tags among the possible tags:
    "ENT", "NEUT" and "CONT".

    Parameters
    -----------
    llm_outputs_with_inputs : Dict
        Passed inputs with a list of tags and explanations for each input statement
        stored in the "inference" key.

    Returns
    ------------
    InferenceScore
        Returns the inference and an explanation of how the inference was obtained.
        If the LLM output is incorrect, the inference is 0.
    """
    try:
        outputs = llm_outputs_with_inputs["inference"]
        passed_statements = llm_outputs_with_inputs["statements"]
        list_tags = [d["tag"] for d in outputs]
        inference = float(np.mean([tag == "ENT" for tag in list_tags]))
        assert len(outputs) == len(passed_statements)
        for d, s in zip(outputs, passed_statements):
            d["statement"] = s
        assert not np.isnan(inference)
        explanation = json.dumps(outputs)
    except (TypeError, KeyError, AssertionError):
        inference = 0.0
        explanation = ""
    return InferenceScore(inference=inference, explanation=explanation)


@chain
def check_if_statements_is_empty(input_: Dict) -> bool:
    assert type(input_) is dict
    return not input_.get("statements")


@chain
def wrap_statements(input_: Dict) -> Dict:
    assert type(input_) is dict
    return {
        "premise": input_["premise"],
        "statements": [
            f"<statement{index + 1}> {statement} </statement{index + 1}>"
            for index, statement in enumerate(input_["statements"])
        ],
        "document": input_["document"],
    }


class LLMInferenceScorer(InferenceScorer):
    """
    The LLMInferenceScorer is designed to calculate
    inference of a hypothesis from a premise using a LLM.
    """

    _chain: RunnableSerializable
    """A chain that contains the core logic, which includes:
    the prompt, model, conversion of output content to JSON,
    and transformation of JSON into InferenceScore."""

    max_concurrency: int
    """Configuration attribute for _chain.batch,
    indicating how many prompts will be sent in parallel."""

    def __init__(
        self,
        model: BaseChatModel,
        max_concurrency: int,
    ):

        self._chain = RunnableBranch(
            (
                check_if_statements_is_empty,
                lambda _: InferenceScore(inference=0.0, explanation=""),
            ),
            RunnablePassthrough.assign(
                inference=wrap_statements | inference_prompt | model | json_to_list
            )
            | returns_to_inference_score,
        )
        self.max_concurrency = max_concurrency

    def get_inference(
        self,
        inference_inputs: List[InferenceInputs],
        show_progress_bar: bool,
    ) -> List[InferenceScore]:
        """
        Method that calls a chain to calculate inference
        of statements from a premise.

        Parameters
        -----------
        inference_inputs : List[InferenceInputs]
            A list of InferenceInputs, where each element includes statements
            for which we want to calculate inference,
            a premise from which we are trying to derive the statements,
            and other additional information for the inference process.

        show_progress_bar : bool
            A flag that controls the display of a progress bar

        Returns
        ------------
        List[InferenceScore]
            Returns the inferences and additionally
            returns an explanation of how the inference was obtained
            for each input.
        """
        with ProgressBarCallback(len(inference_inputs), show_progress_bar) as cb:
            returns = self._chain.batch(
                [
                    {
                        "premise": batch_element.premise,
                        "statements": batch_element.statements,
                        "document": batch_element.document_name.strip(),
                    }
                    for batch_element in inference_inputs
                ],
                config={"callbacks": [cb], "max_concurrency": self.max_concurrency},
            )
        assert isinstance(returns, list)
        return returns



from dataclasses import dataclass, fields
from typing import List, Union

from aidial_rag_eval.types import Answer, GroundTruthAnswer, Text

TextSegment = str
JoinedContext = Text

Premise = Union[JoinedContext, Answer, GroundTruthAnswer]
Hypothesis = Union[Answer, GroundTruthAnswer]
HypothesisSegment = TextSegment
Statement = str
JoinedDocumentsName = str

MetricBind = str


@dataclass
class InferenceInputs:
    """Input data used for calculating inference"""

    hypothesis_id: int
    premise: Premise
    statements: List[Statement]
    document_name: JoinedDocumentsName


@dataclass
class InferenceScore:
    """Inference score for a hypothesis segment, calculated based on InferenceInputs"""

    inference: float
    explanation: str


@dataclass
class InferenceReturn:
    """Inference for a hypothesis, aggregated results for hypothesis segments"""

    inference: float
    json: str
    highlight: str


# Used for calculating mean and median inferences
inference_column = fields(InferenceReturn)[0].name


@dataclass
class RefusalReturn:
    """Answer refusal calculated for the answer"""

    refusal: float



# flake8: noqa
from langchain_core.prompts import PromptTemplate

statement_template = """
Break down each hypothesis into statements, if hypothesis is complex. Else return hypothesis as a single statement.

A statement is a declarative independent self-contained non-overlapping substring forming a complete sentence derived from the hypothesis.

Single words, signs, numbers, links, etc. are not statements.

Your response must be in JSON format:
```json
{
    "hypothesis_statements": [
        {
            "statements": [
                <<statement1 from the first hypothesis>>,
                <<statement2 from the first hypothesis>>,
                ...
            ]
        },
        {
            "statements": [
                <<statement1 from the second hypothesis>>,
                <<statement2 from the second hypothesis>>,
                ...
            ]
        },
        ...
    ]
}
```

Request:
Hypotheses:
{% for item in hypotheses %}
{{ item }}
{% endfor %}
"""

statement_prompt = PromptTemplate.from_template(
    template=statement_template,
    template_format="jinja2",
)



from abc import ABC, abstractmethod
from typing import List

from aidial_rag_eval.generation.types import HypothesisSegment, Statement


class StatementExtractor(ABC):
    """
    Abstract base class for creating StatementExtractor.

    Input is a list of Hypothesis objects.
    """

    @abstractmethod
    def extract(
        self,
        list_of_hypothesis_segments: List[List[HypothesisSegment]],
        show_progress_bar: bool,
    ) -> List[List[List[Statement]]]:
        pass



import pandas as pd

from aidial_rag_eval.dataframe.merge import merge_ground_truth_and_answers
from aidial_rag_eval.facts.citation import CitationMatcher
from aidial_rag_eval.retrieval.types import Matcher

DEFAULT_MATCHER = CitationMatcher


def match_facts(row: pd.Series, matcher: Matcher) -> pd.Series:
    result_row = matcher.match_facts(row.facts, row.context)
    return pd.Series(result_row._asdict())


def apply_matcher_to_merged_dataframe(
    df_merged: pd.DataFrame,
    matcher: Matcher = DEFAULT_MATCHER,
) -> pd.DataFrame:
    matched_results = df_merged.apply(
        match_facts,
        matcher=matcher,
        axis=1,
        result_type="expand",
    )
    assert isinstance(matched_results, pd.DataFrame)
    return matched_results


def match_facts_dataframe(
    ground_truth: pd.DataFrame,
    answers: pd.DataFrame,
    matcher: Matcher = DEFAULT_MATCHER,
) -> pd.DataFrame:
    df_merged = merge_ground_truth_and_answers(ground_truth, answers)
    return pd.merge(
        df_merged,
        apply_matcher_to_merged_dataframe(df_merged, matcher),
        left_index=True,
        right_index=True,
    )



from aidial_rag_eval.generation.inference import (
    calculate_batch_inference,
    calculate_inference,
)
from aidial_rag_eval.generation.refusal import (
    calculate_batch_refusal,
    calculate_refusal,
)
from aidial_rag_eval.retrieval.metrics import (
    calculate_f1,
    calculate_metrics,
    calculate_mrr,
    calculate_precision,
    calculate_recall,
)

__all__ = [
    "calculate_metrics",
    "calculate_f1",
    "calculate_metrics",
    "calculate_mrr",
    "calculate_precision",
    "calculate_recall",
    "calculate_batch_inference",
    "calculate_inference",
    "calculate_batch_refusal",
    "calculate_refusal",
]



from aidial_rag_eval.generation.metric_binds import metric_bind_keys

CONTEXT_TO_ANSWER_INFERENCE = metric_bind_keys[0]
ANSWER_TO_GROUND_TRUTH_INFERENCE = metric_bind_keys[1]
GROUND_TRUTH_TO_ANSWER_INFERENCE = metric_bind_keys[2]
ANSWER_REFUSAL = metric_bind_keys[3]
GROUND_TRUTH_REFUSAL = metric_bind_keys[4]

__all__ = [
    "CONTEXT_TO_ANSWER_INFERENCE",
    "ANSWER_TO_GROUND_TRUTH_INFERENCE",
    "GROUND_TRUTH_TO_ANSWER_INFERENCE",
    "ANSWER_REFUSAL",
    "GROUND_TRUTH_REFUSAL",
]



import itertools
import json
from typing import Iterable, List, Optional, Tuple, TypeVar

import numpy as np
from langchain_core.language_models import BaseChatModel

from aidial_rag_eval.generation.models.converters.llm_decontextualization_converter import (
    LLMNoPronounsConverter,
)
from aidial_rag_eval.generation.models.inference_scorers.llm_inference_scorer import (
    LLMInferenceScorer,
)
from aidial_rag_eval.generation.models.statement_extractor.llm_statement_extractor import (
    LLMStatementExtractor,
)
from aidial_rag_eval.generation.types import (
    Hypothesis,
    HypothesisSegment,
    InferenceInputs,
    InferenceReturn,
    InferenceScore,
    JoinedDocumentsName,
    Premise,
    Statement,
)
from aidial_rag_eval.generation.utils.segmented_text import SegmentedText
from aidial_rag_eval.types import Documents, Question


def _join_documents(documents: Documents) -> JoinedDocumentsName:
    return " ; ".join(documents)


def _make_inference_task_inputs(
    premises: List[Premise],
    statements: List[List[List[Statement]]],
    document_names: List[JoinedDocumentsName],
) -> List[InferenceInputs]:
    """
    The function collects input data for the inference task.

    Parameters
    -----------
    premises : List[str]
        A list of premises from which we want to derive hypotheses in pairs.

    statements : List[List[List[Statement]]]
        A deeply nested list of statements, where the outermost list corresponds to
        different hypotheses, the next level represents the segmentation of each
        hypothesis into hypothesis segments, and the innermost list breaks each
        hypothesis segment down into individual statements.

    document_names: List[str]
        A list of document names used as additional information for the inference task.

    Returns
    ------------
    List[InferenceInputs]
        A list that has as many items as there are innermost lists of statements,
        each innermost statement list is paired with the premise, document names
        and the ID of its origin hypothesis.
    """
    inference_inputs = list(
        itertools.chain.from_iterable(
            [
                [
                    InferenceInputs(
                        hypothesis_id=i,
                        premise=premises[i],
                        statements=list_statements,
                        document_name=document_names[i],
                    )
                    for list_statements in statements[i]
                ]
                for i in range(len(statements))
            ]
        )
    )
    return inference_inputs


T = TypeVar("T")


def _iterable_group_with_key_to_list_group(
    iterable_group_with_key: Tuple[int, Iterable[T]],
) -> List[T]:
    """
    Function that transforms one of the groups obtained from itertools.groupby
    into a more convenient format:
    1) Removes the key used for groupby
    2) Converts the Iterable iterator into a List, preserving the internal objects.

    Parameters
    -----------
    iterable_group_with_key : Tuple[int, Iterable[Any]]
        A group from the results of itertools.groupby.

    Returns
    ------------
    List[Any]
        The same input group, but without the key and in List format.
    """
    return [pair for pair in iterable_group_with_key[1]]


def _grouped_data_item_to_json(
    grouped_data_item: List[Tuple[InferenceInputs, InferenceScore]],
    segmented_text: SegmentedText,
) -> str:
    """
    Function that aggregates the inference results of segments
    for the same hypothesis in JSON format.

    Parameters
    -----------
    grouped_data_item : List[Tuple[InferenceInputs, InferenceScore]]
        Inference results of segments for the same hypothesis.

    segmented_text : SegmentedText
        Segmented hypothesis containing both segments and delimiters
        for reconstructing the original text.

    Returns
    ------------
    str
        JSON string of the inference for the hypothesis.
    """
    return json.dumps(
        [
            {
                "inference": inference_score.inference,
                "hypothesis": segment,
                "premise": [inference_input.premise],
                "explanation": inference_score.explanation,
            }
            for (inference_input, inference_score), segment in zip(
                grouped_data_item, segmented_text.segments
            )
        ]
    )


def _grouped_data_item_to_highlight(
    grouped_data_item: List[Tuple[InferenceInputs, InferenceScore]],
    segmented_text: SegmentedText,
) -> str:
    """
    Function that converts inference results of segments from the same
    hypothesis into a JSON format for text highlighting.

    Parameters
    -----------
    grouped_data_item : List[Tuple[InferenceInputs, InferenceScore]]
        Inference results of segments for the same hypothesis.

    segmented_text : SegmentedText
        Segmented hypothesis containing both segments and delimiters
        for reconstructing the original text.

    Returns
    ------------
    str
        JSON string of highlights intended for coloring segments of the hypothesis.
    """
    highlight = {"corpus": []}
    for (_, inference_score), segment, delimiter in zip(
        grouped_data_item, segmented_text.segments, segmented_text.delimiters + [""]
    ):
        highlight["corpus"].append(
            {
                "text": segment,
                "score": inference_score.inference - 1,
                "title": inference_score.inference,
            }
        )
        highlight["corpus"].append({"text": delimiter, "score": 0.0})
    return json.dumps(highlight)


def segment_hypotheses(
    hypotheses: List[Hypothesis],
    llm: BaseChatModel,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> List[SegmentedText]:
    """
    Function that segments hypotheses into hypothesis segments(roughly into
    sentences), and then removes pronouns using LLM.

    Parameters
    -----------

    hypotheses : List[str]
        The text of the hypothesis.

    llm : BaseChatModel
        The Langchain chat model used for calculating inference.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        Whether to display a progress bar during LLM requests.

    Returns
    ------------
    List[SegmentedText]
        List of hypothesis segments with delimiters.
    """
    converter = LLMNoPronounsConverter(
        model=llm,
        max_concurrency=max_concurrency,
    )

    segmented_hypotheses = [
        SegmentedText.from_text(text=hypothesis) for hypothesis in hypotheses
    ]
    if show_progress_bar:
        print("Converting hypothesis...")
    converter.transform_texts(segmented_hypotheses, show_progress_bar)
    return segmented_hypotheses


def extract_statements(
    list_of_hypothesis_segments: List[List[HypothesisSegment]],
    llm: BaseChatModel,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> List[List[List[Statement]]]:
    """
    Function that extracts statements from each hypothesis segment.
    Hypothesis segments of the inner list are grouped together and
    fed into the prompt.

    Parameters
    -----------

    list_of_hypothesis_segments : List[List[HypothesisSegment]]
        Nested list of hypothesis segments.

    llm : BaseChatModel
        The Langchain chat model used for calculating inference.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        Whether to display a progress bar during LLM requests.

    Returns
    ------------
    List[List[List[Statement]]]
        A deeply nested list of statements, where the outermost list corresponds to
        different hypotheses, the next level corresponds to the segmentation of each
        hypothesis into hypothesis segments, and the innermost list breaks each
        hypothesis segment down into individual statements.
    """
    extractor = LLMStatementExtractor(
        model=llm,
        max_concurrency=max_concurrency,
    )
    if show_progress_bar:
        print("Extracting statements...")
    statements = extractor.extract(
        list_of_hypothesis_segments,
        show_progress_bar,
    )
    return statements


def infer_statements(
    premises: List[Premise],
    statements: List[List[List[Statement]]],
    llm: BaseChatModel,
    questions: Optional[List[Question]] = None,
    list_documents: Optional[List[Documents]] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> List[List[Tuple[InferenceInputs, InferenceScore]]]:
    """
    Function that infers statements.
    Statements of the innermost list are grouped together and
    fed into the prompt.

    Parameters
    -----------

    premises : List[str]
        The text of the premise from which the hypothesis will be inferred.

    statements : List[List[List[Statement]]]
        A deeply nested list of statements, where the outermost list corresponds to
        different hypotheses, the next level corresponds to the segmentation of each
        hypothesis into hypothesis segments, and the innermost list breaks each
        hypothesis segment down into individual statements.

    llm : BaseChatModel
        The Langchain chat model used for calculating inference.

    questions : List[str], optional, default=None
        A questions related to the inference process as a part of the premise.

    list_documents : List[List[str]], optional, default=None
        A list of document names that used
        in the inference process as a part of the premises.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        Whether to display a progress bar during LLM requests.

    Returns
    ------------
    List[List[Tuple[InferenceInputs, InferenceScore]]]
        A nested list of inputs and outputs of the inference step grouped by hypothesis
        segment.
    """
    scorer = LLMInferenceScorer(
        model=llm,
        max_concurrency=max_concurrency,
    )
    if list_documents is None:
        document_names: List[JoinedDocumentsName] = [""] * len(premises)
    else:
        document_names = [_join_documents(docs) for docs in list_documents]
    if questions is not None:
        segmented_questions = [
            SegmentedText.from_text(text=question) for question in questions
        ]
        premises = [
            question_split.segments[-1] + "\n" + premise
            for question_split, premise in zip(segmented_questions, premises)
        ]
    inference_inputs = _make_inference_task_inputs(
        premises,
        statements,
        document_names,
    )
    if show_progress_bar:
        print("Getting inference...")
    inference_scores = scorer.get_inference(
        inference_inputs,
        show_progress_bar,
    )

    iterable_groups_with_id = itertools.groupby(
        zip(inference_inputs, inference_scores), lambda x: x[0].hypothesis_id
    )
    return list(map(_iterable_group_with_key_to_list_group, iterable_groups_with_id))


def calculate_batch_inference(
    premises: List[Premise],
    hypotheses: List[Hypothesis],
    llm: BaseChatModel,
    questions: Optional[List[Question]] = None,
    list_documents: Optional[List[Documents]] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> List[InferenceReturn]:
    """
    Calculates pairwise the inference of a hypotheses from a premises.

    Parameters
    -----------

    premises : List[str]
        The text of the premise from which the hypothesis will be inferred.

    hypotheses : List[str]
        The text of the hypothesis.

    llm : BaseChatModel
        The Langchain chat model used for calculating inference.

    questions : List[str], optional, default=None
        A questions related to the inference process as a part of the premise.

    list_documents : List[List[str]], optional, default=None
        A list of document names that used
        in the inference process as a part of the premises.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        Whether to display a progress bar during LLM requests.

    Returns
    ------------
    List[InferenceReturn]
        Returns the list of inference,
        along with a JSON strings that explains how the inference was derived and
        highlights strings used for highlighting each segment of each hypothesis.
    """

    segmented_hypotheses: List[SegmentedText] = segment_hypotheses(
        hypotheses=hypotheses,
        llm=llm,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    statements: List[List[List[Statement]]] = extract_statements(
        list_of_hypothesis_segments=[
            segmented_hypothesis.segments
            for segmented_hypothesis in segmented_hypotheses
        ],
        llm=llm,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    grouped_data_list: List[List[Tuple[InferenceInputs, InferenceScore]]] = (
        infer_statements(
            premises=premises,
            statements=statements,
            llm=llm,
            questions=questions,
            list_documents=list_documents,
            max_concurrency=max_concurrency,
            show_progress_bar=show_progress_bar,
        )
    )

    aggregated_inferences = map(
        lambda grouped_data_item: float(
            np.mean(
                [inference_score.inference for _, inference_score in grouped_data_item]
            )
        ),
        grouped_data_list,
    )

    aggregated_jsons = itertools.starmap(
        _grouped_data_item_to_json,
        zip(grouped_data_list, segmented_hypotheses),
    )
    highlights = itertools.starmap(
        _grouped_data_item_to_highlight, zip(grouped_data_list, segmented_hypotheses)
    )
    inference_returns = [
        InferenceReturn(inference=inference, json=js, highlight=highlight)
        for inference, js, highlight in zip(
            aggregated_inferences, aggregated_jsons, highlights
        )
    ]
    return inference_returns


def calculate_inference(
    premise: Premise,
    hypothesis: Hypothesis,
    llm: BaseChatModel,
    question: Optional[Question] = None,
    documents: Optional[Documents] = None,
    max_concurrency: int = 8,
    show_progress_bar: bool = True,
) -> InferenceReturn:
    """
    Calculates the inference of a hypothesis from a premise.

    Parameters
    -----------

    premise : str
        The text of the premise from which the hypothesis will be inferred.

    hypothesis : str
        The text of the hypothesis.

    llm : BaseChatModel
        The Langchain chat model used for calculating inference.

    question : str, optional, default=None
        A question related to the inference process as a part of the premise.

    documents : List[str], optional, default=None
        A document names that used in the inference process  as a part of the premise.

    max_concurrency : int, default=8
        The maximum number of concurrent requests to the LLM.

    show_progress_bar : bool, default=True
        Whether to display a progress bar during LLM requests.

    Returns
    ------------
    InferenceReturn
        Returns the inference,
        along with a JSON string that explains how the inference was derived and
        highlights string used for highlighting each segment of the hypothesis.
    """
    questions = None if question is None else [question]
    list_documents = None if documents is None else [documents]
    inference_returns = calculate_batch_inference(
        premises=[premise],
        hypotheses=[hypothesis],
        llm=llm,
        questions=questions,
        list_documents=list_documents,
        max_concurrency=max_concurrency,
        show_progress_bar=show_progress_bar,
    )
    return inference_returns[0]



from typing import NamedTuple, Protocol, runtime_checkable

import numpy as np
import numpy.typing as npt

from aidial_rag_eval.types import ContextChunk, FactType

FactsRanks = np.ndarray
ContextRelevance = np.ndarray
ContextHighlight = npt.NDArray[np.str_]


class FactMatchResult(NamedTuple):
    facts_ranks: FactsRanks
    context_relevance: ContextRelevance
    context_highlight: ContextHighlight


@runtime_checkable
class Matcher(Protocol[FactType]):
    @staticmethod
    def match_facts(  # noqa: E704
        facts: list[FactType], context: list[ContextChunk]
    ) -> FactMatchResult: ...



import json
from typing import Tuple

import numpy as np

from aidial_rag_eval.retrieval.types import (
    ContextChunk,
    ContextRelevance,
    FactMatchResult,
    FactsRanks,
)
from aidial_rag_eval.types import Context, Facts


def highlight_chunk(
    context_chunk: ContextChunk, facts_in_chunk: list[Tuple[int, Tuple[int, int]]]
) -> str:
    chunk_highlight = []

    fact_events = []
    for j, (start, end) in facts_in_chunk:
        fact_events.append((start, False, j))
        fact_events.append((end, True, j))
    fact_events.sort()

    open_facts = set()
    prev_pos = 0
    for pos, is_end, j in fact_events:
        if prev_pos < pos:
            chunk_highlight.append(
                {"text": context_chunk[prev_pos:pos], "facts": list(open_facts)}
            )
        prev_pos = pos

        if is_end:
            open_facts.remove(j)
        else:
            open_facts.add(j)
    if prev_pos < len(context_chunk):
        chunk_highlight.append(
            {"text": context_chunk[prev_pos:], "facts": list(open_facts)}
        )

    return json.dumps({"match": chunk_highlight})


def match_str_facts(
    facts: Facts[str], context: Context, match_fact_func
) -> FactMatchResult:
    facts_ranks: FactsRanks = np.full(len(facts), -1, dtype=int)
    context_relevance: ContextRelevance = np.zeros(len(context), dtype=int)
    context_highlight = []

    for i, c in enumerate(context):
        facts_in_chunk = []
        for j, fact in enumerate(facts):
            match = match_fact_func(fact, c)
            if match is not None:
                if facts_ranks[j] == -1:
                    facts_ranks[j] = i
                else:
                    facts_ranks[j] = min(facts_ranks[j], i)
                facts_in_chunk.append((j, match))
        context_relevance[i] = len(facts_in_chunk)
        context_highlight.append(highlight_chunk(c, facts_in_chunk))

    return FactMatchResult(
        facts_ranks=facts_ranks,
        context_relevance=context_relevance,
        context_highlight=np.array(context_highlight, dtype=object),
    )



# flake8: noqa
from langchain_core.prompts import PromptTemplate

inference_template = """
Natural language inference is the task of determining whether the hypothesis is an entailment, contradiction, or neutral with respect to the premise.
A hypothesis is a list of statements provided below. 

{% if document %}
The name of the document from which the premise was derived is also provided.
{% endif %}

A statement is considered an entailment if it logically follows from the premise.
A statement is considered a contradiction if it is logically inconsistent with the premise.
Else a statement is considered a neutral.
Important: Don't reject entailment just because of minor extra details - if the main meaning holds, it's still entailment.

For each statement:
Provide a brief short(1 sentences) explanation of whether the statement is an entailment, contradiction or neutral with respect to the premise.
Assign tags based on your explanation: "ENT" for entailment, "CONT" for contradiction, "NEUT" for neutral or if none of the above tags apply.

Format your response in JSON. You must return only JSON.

For example, if the premise is "I am a biology graduate and I work at a tech company." and the list of statements is ["I am a graduate.", "I work at a hospital."] your response should be:

```json
{
    "statement_inference": [
        {
            "explanation": "It is true that I am a graduate",
            "tag": "ENT"
        },
        {
            "explanation": "Premise states I work at a tech company, not a hospital.",
            "tag": "CONT"
        }
    ]
}
```

Your response must be in JSON format:
```json
{
    "statement_inference": [
        {
            "explanation": <<explanation>>,
            "tag": <<"ENT" or "CONT" or "NEUT">>
        },
        ...
    ]
}
```
Request:

{% if document %}
<document_name>
{{ document }}
</document_name>
{% endif %}
<premise>
{{ premise }}
</premise>

List of statements:
{% for item in statements %}
{{ item }}
{% endfor %}
"""

inference_prompt = PromptTemplate.from_template(
    template=inference_template,
    template_format="jinja2",
)



from typing import Any

from langchain_core.callbacks import BaseCallbackHandler
from langchain_core.outputs import LLMResult
from tqdm.autonotebook import tqdm


class ProgressBarCallback(BaseCallbackHandler):
    def __init__(self, total: int, show_progress_bar: bool):
        super().__init__()
        self.progress_bar = tqdm(total=total, disable=not show_progress_bar)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        self.progress_bar.update(1)

    def __enter__(self):
        self.progress_bar.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.progress_bar.__exit__(exc_type, exc_value, exc_traceback)

    def __del__(self):
        self.progress_bar.__del__()



# flake8: noqa
from langchain_core.prompts import PromptTemplate

refusal_template = """
Answer Refusal task is to determine if an answer should be tagged as a refusal to answer based on specific criteria.
The criteria include identifying if the answer explicitly states something is wrong with the question, request, or premise, or if it indicates a lack of information, irrelevance, or refusal to answer.

Single words, signs, numbers, links, etc. are not considered refusal to answer.

Lack of information can be formulated in different ways, pay attention to the list of synonyms.
Synonymous series:
1) premise, context, document, information.
2) hypothesis, answer.

Tagging guidelines:
- Use "REJ" if the answer is answer refusal, else tag it "ANS".

Example:
An explicit statement: "There is no answer to this question." should be tagged "REJ".
A statement that is not explicit: "The answer to this question is yes." should be tagged "ANS".

Format your response in JSON:
```json
{
    "tags": [
        <<"REJ" or "ANS">>,
        <<"REJ" or "ANS">>,
        ...
    ]
}
```

Each answer from the list of answers corresponds to a tag in your response.
The first answer corresponds to the first tag, the second corresponds to the second.
The number of tags must be the same as the number of answers in the answer list.
Each answer, even meaningless, must have it's own tag, if you don't know how to tag it, than just leave "ANS" tag for it.

Request:
List of answers:
{% for item in answers %}
{{ item }}
{% endfor %}"""

refusal_prompt = PromptTemplate.from_template(
    template=refusal_template,
    template_format="jinja2",
)
